from qtpy.QtWidgets import QHBoxLayout, QToolBar, QWidget

from bec_widgets.utils.toolbars.actions import NoCheckDelegate, ToolBarAction
from bec_widgets.widgets.control.device_input.base_classes.device_input_base import BECDeviceFilter
from bec_widgets.widgets.control.device_input.device_combobox.device_combobox import DeviceComboBox


class MotorSelectionAction(ToolBarAction):
    def __init__(self, parent=None):
        super().__init__(icon_path=None, tooltip=None, checkable=False)
        self.motor_x = DeviceComboBox(parent=parent, device_filter=[BECDeviceFilter.POSITIONER])
        self.motor_x.addItem("", None)
        self.motor_x.setCurrentText("")
        self.motor_x.setToolTip("Select Motor X")
        self.motor_x.setItemDelegate(NoCheckDelegate(self.motor_x))
        self.motor_y = DeviceComboBox(parent=parent, device_filter=[BECDeviceFilter.POSITIONER])
        self.motor_y.addItem("", None)
        self.motor_y.setCurrentText("")
        self.motor_y.setToolTip("Select Motor Y")
        self.motor_y.setItemDelegate(NoCheckDelegate(self.motor_y))

        self.container = QWidget(parent)
        layout = QHBoxLayout(self.container)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.motor_x)
        layout.addWidget(self.motor_y)
        self.container.setLayout(layout)
        self.action = self.container

    def add_to_toolbar(self, toolbar: QToolBar, target: QWidget):
        """
        Adds the widget to the toolbar.

        Args:
            toolbar (QToolBar): The toolbar to add the widget to.
            target (QWidget): The target widget for the action.
        """

        toolbar.addWidget(self.container)

    def cleanup(self):
        """
        Cleans up the action, if necessary.
        """
        self.motor_x.close()
        self.motor_x.deleteLater()
        self.motor_y.close()
        self.motor_y.deleteLater()
        self.container.close()
        self.container.deleteLater()



def main():  # pragma: no cover
    from qtpy import PYSIDE6

    if not PYSIDE6:
        print("PYSIDE6 is not available in the environment. Cannot patch designer.")
        return
    from PySide6.QtDesigner import QPyDesignerCustomWidgetCollection

    from bec_widgets.widgets.control.scan_control.scan_control_plugin import ScanControlPlugin

    QPyDesignerCustomWidgetCollection.addCustomWidget(ScanControlPlugin())


if __name__ == "__main__":  # pragma: no cover
    main()



from typing import Literal

import qtmonaco
from qtpy.QtWidgets import QApplication, QVBoxLayout, QWidget

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.colors import get_theme_name


class MonacoWidget(BECWidget, QWidget):
    """
    A simple Monaco editor widget
    """

    PLUGIN = True
    ICON_NAME = "code"
    USER_ACCESS = [
        "set_text",
        "get_text",
        "set_language",
        "get_language",
        "set_theme",
        "get_theme",
        "set_readonly",
        "set_cursor",
        "current_cursor",
        "set_minimap_enabled",
    ]

    def __init__(self, parent=None, config=None, client=None, gui_id=None, **kwargs):
        super().__init__(
            parent=parent, client=client, gui_id=gui_id, config=config, theme_update=True, **kwargs
        )
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.editor = qtmonaco.Monaco(self)
        layout.addWidget(self.editor)
        self.setLayout(layout)
        self.editor.initialized.connect(self.apply_theme)

    def apply_theme(self, theme: str | None = None) -> None:
        """
        Apply the current theme to the Monaco editor.

        Args:
            theme (str, optional): The theme to apply. If None, the current theme will be used.
        """
        if theme is None:
            theme = get_theme_name()
        editor_theme = "vs" if theme == "light" else "vs-dark"
        self.set_theme(editor_theme)

    def set_text(self, text: str) -> None:
        """
        Set the text in the Monaco editor.

        Args:
            text (str): The text to set in the editor.
        """
        self.editor.set_text(text)

    def get_text(self) -> str:
        """
        Get the current text from the Monaco editor.
        """
        return self.editor.get_text()

    def set_cursor(
        self,
        line: int,
        column: int = 1,
        move_to_position: Literal[None, "center", "top", "position"] = None,
    ) -> None:
        """
        Set the cursor position in the Monaco editor.

        Args:
            line (int): Line number (1-based).
            column (int): Column number (1-based), defaults to 1.
            move_to_position (Literal[None, "center", "top", "position"], optional): Position to move the cursor to.
        """
        self.editor.set_cursor(line, column, move_to_position)

    def current_cursor(self) -> dict[str, int]:
        """
        Get the current cursor position in the Monaco editor.

        Returns:
            dict[str, int]: A dictionary with keys 'line' and 'column' representing the cursor position.
        """
        return self.editor.current_cursor

    def set_language(self, language: str) -> None:
        """
        Set the programming language for syntax highlighting in the Monaco editor.

        Args:
            language (str): The programming language to set (e.g., "python", "javascript").
        """
        self.editor.set_language(language)

    def get_language(self) -> str:
        """
        Get the current programming language set in the Monaco editor.
        """
        return self.editor.get_language()

    def set_readonly(self, read_only: bool) -> None:
        """
        Set the Monaco editor to read-only mode.

        Args:
            read_only (bool): If True, the editor will be read-only.
        """
        self.editor.set_readonly(read_only)

    def set_theme(self, theme: str) -> None:
        """
        Set the theme for the Monaco editor.

        Args:
            theme (str): The theme to set (e.g., "vs-dark", "light").
        """
        self.editor.set_theme(theme)

    def get_theme(self) -> str:
        """
        Get the current theme of the Monaco editor.
        """
        return self.editor.get_theme()

    def set_minimap_enabled(self, enabled: bool) -> None:
        """
        Enable or disable the minimap in the Monaco editor.

        Args:
            enabled (bool): If True, the minimap will be enabled; otherwise, it will be disabled.
        """
        self.editor.set_minimap_enabled(enabled)

    def set_highlighted_lines(self, start_line: int, end_line: int) -> None:
        """
        Highlight a range of lines in the Monaco editor.

        Args:
            start_line (int): The starting line number (1-based).
            end_line (int): The ending line number (1-based).
        """
        self.editor.set_highlighted_lines(start_line, end_line)

    def clear_highlighted_lines(self) -> None:
        """
        Clear any highlighted lines in the Monaco editor.
        """
        self.editor.clear_highlighted_lines()


if __name__ == "__main__":  # pragma: no cover
    qapp = QApplication([])
    widget = MonacoWidget()
    # set the default size
    widget.resize(800, 600)
    widget.set_language("python")
    widget.set_theme("vs-dark")
    widget.editor.set_minimap_enabled(False)
    widget.set_text(
        """
import numpy as np
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bec_lib.devicemanager import DeviceContainer
    from bec_lib.scans import Scans
    dev: DeviceContainer
    scans: Scans

#######################################
########## User Script #####################
#######################################

# This is a comment
def hello_world():
    print("Hello, world!")
            """
    )
    widget.set_highlighted_lines(1, 3)
    widget.show()
    qapp.exec_()



from __future__ import annotations

from abc import abstractmethod

from qtpy.QtCore import QObject


class BundleConnection(QObject):
    bundle_name: str

    @abstractmethod
    def connect(self):
        """
        Connects the bundle to the target widget or application.
        This method should be implemented by subclasses to define how the bundle interacts with the target.
        """

    @abstractmethod
    def disconnect(self):
        """
        Disconnects the bundle from the target widget or application.
        This method should be implemented by subclasses to define how to clean up connections.
        """



from __future__ import annotations

from typing import TYPE_CHECKING

from bec_lib.callback_handler import EventType
from bec_lib.logger import bec_logger
from bec_lib.messages import ScanHistoryMessage
from qtpy import QtCore, QtGui, QtWidgets

from bec_widgets.utils.bec_widget import BECWidget, ConnectionConfig
from bec_widgets.utils.colors import get_accent_colors
from bec_widgets.utils.error_popups import SafeSlot

if TYPE_CHECKING:
    from bec_lib.client import BECClient


logger = bec_logger.logger


class BECHistoryManager(QtCore.QObject):
    """History manager for scan history operations. This class
    is responsible for emitting signals when the scan history is updated.
    """

    # ScanHistoryMessage.model_dump() (dict)
    scan_history_updated = QtCore.Signal(dict)

    def __init__(self, parent, client: BECClient):
        super().__init__(parent)
        self.client = client
        self._cb_id = self.client.callbacks.register(
            event_type=EventType.SCAN_HISTORY_UPDATE, callback=self._on_scan_history_update
        )

    def refresh_scan_history(self) -> None:
        """Refresh the scan history from the client."""
        for scan_id in self.client.history._scan_ids:  # pylint: disable=protected-access
            history_msg = self.client.history._scan_data.get(scan_id, None)
            if history_msg is None:
                logger.info(f"Scan history message for scan_id {scan_id} not found.")
                continue
            self.scan_history_updated.emit(history_msg.model_dump())

    def _on_scan_history_update(self, history_msg: ScanHistoryMessage) -> None:
        """Handle scan history updates from the client."""
        self.scan_history_updated.emit(history_msg.model_dump())

    def cleanup(self) -> None:
        """Clean up the manager by disconnecting callbacks."""
        self.client.callbacks.remove(self._cb_id)
        self.scan_history_updated.disconnect()


class ScanHistoryView(BECWidget, QtWidgets.QTreeWidget):
    """ScanHistoryTree is a widget that displays the scan history in a tree format."""

    RPC = False
    PLUGIN = False

    # ScanHistoryMessage.content, ScanHistoryMessage.metadata
    scan_selected = QtCore.Signal(dict, dict)
    no_scan_selected = QtCore.Signal()

    def __init__(
        self,
        parent: QtWidgets.QWidget = None,
        client=None,
        config: ConnectionConfig = None,
        gui_id: str = None,
        max_length: int = 100,
        theme_update: bool = True,
        **kwargs,
    ):
        super().__init__(
            parent=parent,
            client=client,
            config=config,
            gui_id=gui_id,
            theme_update=theme_update,
            **kwargs,
        )
        colors = get_accent_colors()
        self.status_colors = {
            "closed": colors.success,
            "halted": colors.warning,
            "aborted": colors.emergency,
        }
        # self.status_colors = {"closed": "#00e676", "halted": "#ffca28", "aborted": "#ff5252"}
        self.column_header = ["Scan Nr", "Scan Name", "Status"]
        self.scan_history: list[ScanHistoryMessage] = []  # newest at index 0
        self.max_length = max_length  # Maximum number of scan history entries to keep
        self.bec_scan_history_manager = BECHistoryManager(parent=self, client=self.client)
        self._set_policies()
        self.apply_theme()
        self.currentItemChanged.connect(self._current_item_changed)
        header = self.header()
        header.setToolTip(f"Last {self.max_length} scans in history.")
        self.bec_scan_history_manager.scan_history_updated.connect(self.update_history)
        self.refresh()

    def _set_policies(self):
        """Set the policies for the tree widget."""
        self.setColumnCount(len(self.column_header))
        self.setHeaderLabels(self.column_header)
        self.setRootIsDecorated(False)  # allow expand arrow for per‑scan details
        self.setUniformRowHeights(True)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setAlternatingRowColors(True)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.setIndentation(12)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setAnimated(True)

        header = self.header()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
        for column in range(1, self.columnCount()):
            header.setSectionResizeMode(column, QtWidgets.QHeaderView.ResizeMode.Stretch)

    def apply_theme(self, theme: str | None = None):
        """Apply the theme to the widget."""
        colors = get_accent_colors()
        self.status_colors = {
            "closed": colors.success,
            "halted": colors.warning,
            "aborted": colors.emergency,
        }
        self.repaint()

    def _current_item_changed(
        self, current: QtWidgets.QTreeWidgetItem, previous: QtWidgets.QTreeWidgetItem
    ):
        """
        Handle current item change events in the tree widget.

        Args:
            current (QtWidgets.QTreeWidgetItem): The currently selected item.
            previous (QtWidgets.QTreeWidgetItem): The previously selected item.
        """
        if not current:
            return
        index = self.indexOfTopLevelItem(current)
        self.scan_selected.emit(self.scan_history[index].content, self.scan_history[index].metadata)

    @SafeSlot()
    def refresh(self):
        """Refresh the scan history view."""
        while len(self.scan_history) > 0:
            self.remove_scan(index=0)
        self.bec_scan_history_manager.refresh_scan_history()

    @SafeSlot(dict)
    def update_history(self, msg_dump: dict):
        """Update the scan history with new scan data."""
        msg = ScanHistoryMessage(**msg_dump)
        self.add_scan(msg)
        self.ensure_history_max_length()

    def ensure_history_max_length(self) -> None:
        """
        Method to ensure the scan history does not exceed the maximum length.
        If the length exceeds the maximum, it removes the oldest entry.
        This is called after adding a new scan to the history.
        """
        while len(self.scan_history) > self.max_length:
            logger.warning(
                f"Removing oldest scan history entry to maintain max length of {self.max_length}."
            )
            self.remove_scan(index=-1)

    def add_scan(self, msg: ScanHistoryMessage):
        """
        Add a scan entry to the tree widget.

        Args:
            msg (ScanHistoryMessage): The scan history message containing scan details.
        """
        if msg.stored_data_info is None:
            logger.info(
                f"Old scan history entry fo scan {msg.scan_id} without stored_data_info, skipping."
            )
            return
        if msg in self.scan_history:
            logger.info(f"Scan {msg.scan_id} already in history, skipping.")
            return
        self.scan_history.insert(0, msg)
        tree_item = QtWidgets.QTreeWidgetItem([str(msg.scan_number), msg.scan_name, ""])
        color = QtGui.QColor(self.status_colors.get(msg.exit_status, "#b0bec5"))
        pix = QtGui.QPixmap(10, 10)
        pix.fill(QtCore.Qt.transparent)
        with QtGui.QPainter(pix) as p:
            p.setRenderHint(QtGui.QPainter.Antialiasing)
            p.setPen(QtCore.Qt.NoPen)
            p.setBrush(color)
            p.drawEllipse(0, 0, 10, 10)
        tree_item.setIcon(2, QtGui.QIcon(pix))
        tree_item.setForeground(2, QtGui.QBrush(color))
        for col in range(tree_item.columnCount()):
            tree_item.setToolTip(col, f"Status: {msg.exit_status}")
        self.insertTopLevelItem(0, tree_item)
        tree_item.setExpanded(False)

    def remove_scan(self, index: int):
        """
        Remove a scan entry from the tree widget.
        We supoprt negative indexing where -1, -2, etc.

        Args:
            index (int): The index of the scan entry to remove.
        """
        if index < 0:
            index = len(self.scan_history) + index
        try:
            msg = self.scan_history.pop(index)
            self.no_scan_selected.emit()
        except IndexError:
            logger.warning(f"Invalid index {index} for removing scan entry from history.")
            return
        self.takeTopLevelItem(index)

    def cleanup(self):
        """Cleanup the widget"""
        self.bec_scan_history_manager.cleanup()
        super().cleanup()


if __name__ == "__main__":  # pragma: no cover
    # pylint: disable=import-outside-toplevel

    from bec_widgets.widgets.services.scan_history_browser.components import (
        ScanHistoryDeviceViewer,
        ScanHistoryMetadataViewer,
    )
    from bec_widgets.widgets.utility.visual.dark_mode_button.dark_mode_button import DarkModeButton

    app = QtWidgets.QApplication([])

    main_window = QtWidgets.QMainWindow()
    central_widget = QtWidgets.QWidget()
    button = DarkModeButton()
    layout = QtWidgets.QVBoxLayout(central_widget)
    main_window.setCentralWidget(central_widget)

    # Create a ScanHistoryBrowser instance
    browser = ScanHistoryView()

    # Create a ScanHistoryView instance
    view = ScanHistoryMetadataViewer()
    device_viewer = ScanHistoryDeviceViewer()

    layout.addWidget(button)
    layout.addWidget(browser)
    layout.addWidget(view)
    layout.addWidget(device_viewer)
    browser.scan_selected.connect(view.update_view)
    browser.scan_selected.connect(device_viewer.update_devices_from_scan_history)
    browser.no_scan_selected.connect(view.clear_view)
    browser.no_scan_selected.connect(device_viewer.clear_view)

    main_window.show()
    app.exec_()



def main():  # pragma: no cover
    from qtpy import PYSIDE6

    if not PYSIDE6:
        print("PYSIDE6 is not available in the environment. Cannot patch designer.")
        return
    from PySide6.QtDesigner import QPyDesignerCustomWidgetCollection

    from bec_widgets.widgets.control.device_input.signal_line_edit.signal_line_edit_plugin import (
        SignalLineEditPlugin,
    )

    QPyDesignerCustomWidgetCollection.addCustomWidget(SignalLineEditPlugin())


if __name__ == "__main__":  # pragma: no cover
    main()



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, DefaultDict
from weakref import ReferenceType

import louie
from bec_lib.logger import bec_logger
from pydantic import BaseModel

from bec_widgets.utils.toolbars.actions import SeparatorAction, ToolBarAction

if TYPE_CHECKING:
    from bec_widgets.utils.toolbars.connections import BundleConnection
    from bec_widgets.utils.toolbars.toolbar import ModularToolBar

logger = bec_logger.logger


class ActionInfo(BaseModel):
    action: ToolBarAction
    toolbar_bundle: ToolbarBundle | None = None

    model_config = {"arbitrary_types_allowed": True}


class ToolbarComponents:
    def __init__(self, toolbar: ModularToolBar):
        """
        Initializes the toolbar components.

        Args:
            toolbar (ModularToolBar): The toolbar to which the components will be added.
        """
        self.toolbar = toolbar

        self._components: dict[str, ActionInfo] = {}
        self.add("separator", SeparatorAction())

    def add(self, name: str, component: ToolBarAction):
        """
        Adds a component to the toolbar.

        Args:
            component (ToolBarAction): The component to add.
        """
        if name in self._components:
            raise ValueError(f"Component with name '{name}' already exists.")
        self._components[name] = ActionInfo(action=component, toolbar_bundle=None)

    def add_safe(self, name: str, component: ToolBarAction):
        """
        Adds a component to the toolbar, ensuring it does not already exist.

        Args:
            name (str): The name of the component.
            component (ToolBarAction): The component to add.
        """
        if self.exists(name):
            logger.info(f"Component with name '{name}' already exists. Skipping addition.")
            return
        self.add(name, component)

    def exists(self, name: str) -> bool:
        """
        Checks if a component exists in the toolbar.

        Args:
            name (str): The name of the component to check.

        Returns:
            bool: True if the component exists, False otherwise.
        """
        return name in self._components

    def get_action_reference(self, name: str) -> ReferenceType[ToolBarAction]:
        """
        Retrieves a component by name.

        Args:
            name (str): The name of the component to retrieve.

        """
        if not self.exists(name):
            raise KeyError(f"Component with name '{name}' does not exist.")
        return louie.saferef.safe_ref(self._components[name].action)

    def get_action(self, name: str) -> ToolBarAction:
        """
        Retrieves a component by name.

        Args:
            name (str): The name of the component to retrieve.

        Returns:
            ToolBarAction: The action associated with the given name.
        """
        if not self.exists(name):
            raise KeyError(
                f"Component with name '{name}' does not exist. The following components are available: {list(self._components.keys())}"
            )
        return self._components[name].action

    def set_bundle(self, name: str, bundle: ToolbarBundle):
        """
        Sets the bundle for a component.

        Args:
            name (str): The name of the component.
            bundle (ToolbarBundle): The bundle to set.
        """
        if not self.exists(name):
            raise KeyError(f"Component with name '{name}' does not exist.")
        comp = self._components[name]
        if comp.toolbar_bundle is not None:
            logger.info(
                f"Component '{name}' already has a bundle ({comp.toolbar_bundle.name}). Setting it to {bundle.name}."
            )
            comp.toolbar_bundle.bundle_actions.pop(name, None)
        comp.toolbar_bundle = bundle

    def remove_action(self, name: str):
        """
        Removes a component from the toolbar.

        Args:
            name (str): The name of the component to remove.
        """
        if not self.exists(name):
            raise KeyError(f"Action with ID '{name}' does not exist.")
        action_info = self._components.pop(name)
        if action_info.toolbar_bundle:
            action_info.toolbar_bundle.bundle_actions.pop(name, None)
            self.toolbar.refresh()
            action_info.toolbar_bundle = None
        if hasattr(action_info.action, "cleanup"):
            # Call cleanup if the action has a cleanup method
            action_info.action.cleanup()

    def cleanup(self):
        """
        Cleans up the toolbar components by removing all actions and bundles.
        """
        for action_info in self._components.values():
            if hasattr(action_info.action, "cleanup"):
                # Call cleanup if the action has a cleanup method
                action_info.action.cleanup()
        self._components.clear()


class ToolbarBundle:
    def __init__(self, name: str, components: ToolbarComponents):
        """
        Initializes a new bundle component.

        Args:
            bundle_name (str): Unique identifier for the bundle.
        """
        self.name = name
        self.components = components
        self.bundle_actions: DefaultDict[str, ReferenceType[ToolBarAction]] = defaultdict()
        self._connections: dict[str, BundleConnection] = {}

    def add_action(self, name: str):
        """
        Adds an action to the bundle.

        Args:
            name (str): Unique identifier for the action.
            action (ToolBarAction): The action to add.
        """
        if name in self.bundle_actions:
            raise ValueError(f"Action with name '{name}' already exists in bundle '{self.name}'.")
        if not self.components.exists(name):
            raise ValueError(
                f"Component with name '{name}' does not exist in the toolbar. Please add it first using the `ToolbarComponents.add` method."
            )
        self.bundle_actions[name] = self.components.get_action_reference(name)
        self.components.set_bundle(name, self)

    def remove_action(self, name: str):
        """
        Removes an action from the bundle.

        Args:
            name (str): The name of the action to remove.
        """
        if name not in self.bundle_actions:
            raise KeyError(f"Action with name '{name}' does not exist in bundle '{self.name}'.")
        del self.bundle_actions[name]

    def add_separator(self):
        """
        Adds a separator action to the bundle.
        """
        self.add_action("separator")

    def add_connection(self, name: str, connection):
        """
        Adds a connection to the bundle.

        Args:
            name (str): Unique identifier for the connection.
            connection: The connection to add.
        """
        if name in self._connections:
            raise ValueError(
                f"Connection with name '{name}' already exists in bundle '{self.name}'."
            )
        self._connections[name] = connection

    def remove_connection(self, name: str):
        """
        Removes a connection from the bundle.

        Args:
            name (str): The name of the connection to remove.
        """
        if name not in self._connections:
            raise KeyError(f"Connection with name '{name}' does not exist in bundle '{self.name}'.")
        self._connections[name].disconnect()
        del self._connections[name]

    def get_connection(self, name: str):
        """
        Retrieves a connection by name.

        Args:
            name (str): The name of the connection to retrieve.

        Returns:
            The connection associated with the given name.
        """
        if name not in self._connections:
            raise KeyError(f"Connection with name '{name}' does not exist in bundle '{self.name}'.")
        return self._connections[name]

    def disconnect(self):
        """
        Disconnects all connections in the bundle.
        """
        for connection in self._connections.values():
            connection.disconnect()
        self._connections.clear()



from __future__ import annotations

import os

from bec_lib.endpoints import MessageEndpoints
from qtpy.QtCore import QEasingCurve, QEvent, QPropertyAnimation, QSize, Qt, QTimer
from qtpy.QtGui import QAction, QActionGroup, QIcon
from qtpy.QtWidgets import (
    QApplication,
    QFrame,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QStyle,
    QVBoxLayout,
    QWidget,
)

import bec_widgets
from bec_widgets.utils import UILoader
from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.colors import apply_theme
from bec_widgets.utils.error_popups import SafeSlot
from bec_widgets.utils.widget_io import WidgetHierarchy
from bec_widgets.widgets.containers.main_window.addons.hover_widget import HoverWidget
from bec_widgets.widgets.containers.main_window.addons.scroll_label import ScrollLabel
from bec_widgets.widgets.containers.main_window.addons.web_links import BECWebLinksMixin
from bec_widgets.widgets.progress.scan_progressbar.scan_progressbar import ScanProgressBar

MODULE_PATH = os.path.dirname(bec_widgets.__file__)


class BECMainWindow(BECWidget, QMainWindow):
    RPC = False
    PLUGIN = False
    SCAN_PROGRESS_WIDTH = 100  # px
    STATUS_BAR_WIDGETS_EXPIRE_TIME = 60_000  # milliseconds

    def __init__(
        self,
        parent=None,
        gui_id: str = None,
        client=None,
        window_title: str = "BEC",
        *args,
        **kwargs,
    ):
        super().__init__(parent=parent, gui_id=gui_id, **kwargs)

        self.app = QApplication.instance()
        self.status_bar = self.statusBar()
        self.setWindowTitle(window_title)
        self._init_ui()
        self._connect_to_theme_change()

        # Connections to BEC Notifications
        self.bec_dispatcher.connect_slot(
            self.display_client_message, MessageEndpoints.client_info()
        )

    ################################################################################
    # MainWindow Elements Initialization
    ################################################################################
    def _init_ui(self):

        # Set the icon
        self._init_bec_icon()

        # Set Menu and Status bar
        self._setup_menu_bar()
        self._init_status_bar_widgets()

        # BEC Specific UI
        self.display_app_id()

    def _init_status_bar_widgets(self):
        """
        Prepare the BEC specific widgets in the status bar.
        """

        # Left: App‑ID label
        self._app_id_label = QLabel()
        self._app_id_label.setAlignment(
            Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter
        )
        self.status_bar.addWidget(self._app_id_label)

        # Add a separator after the app ID label
        self._add_separator()

        # Centre: Client‑info label (stretch=1 so it expands)
        self._add_client_info_label()

        # Add scan_progress bar with display logic
        self._add_scan_progress_bar()

    ################################################################################
    # Client message status bar widget helpers

    def _add_client_info_label(self):
        """
        Add a client info label to the status bar.
        This label will display messages from the BEC dispatcher.
        """

        # Scroll label for client info in Status Bar
        self._client_info_label = ScrollLabel(self)
        self._client_info_label.setAlignment(
            Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter
        )
        # Full label used in the hover widget
        self._client_info_label_full = QLabel(self)
        self._client_info_label_full.setAlignment(
            Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter
        )
        # Hover widget to show the full client info label
        self._client_info_hover = HoverWidget(
            self, simple=self._client_info_label, full=self._client_info_label_full
        )
        self.status_bar.addWidget(self._client_info_hover, 1)

        # Timer to automatically clear client messages once they expire
        self._client_info_expire_timer = QTimer(self)
        self._client_info_expire_timer.setSingleShot(True)
        self._client_info_expire_timer.timeout.connect(lambda: self._client_info_label.setText(""))
        self._client_info_expire_timer.timeout.connect(
            lambda: self._client_info_label_full.setText("")
        )

    ################################################################################
    # Progress‑bar helpers
    def _add_scan_progress_bar(self):

        # Setting HoverWidget for the scan progress bar - minimal and full version
        self._scan_progress_bar_simple = ScanProgressBar(self, one_line_design=True)
        self._scan_progress_bar_simple.show_elapsed_time = False
        self._scan_progress_bar_simple.show_remaining_time = False
        self._scan_progress_bar_simple.show_source_label = False
        self._scan_progress_bar_simple.progressbar.label_template = ""
        self._scan_progress_bar_simple.progressbar.setFixedHeight(8)
        self._scan_progress_bar_simple.progressbar.setFixedWidth(80)
        self._scan_progress_bar_full = ScanProgressBar(self)
        self._scan_progress_hover = HoverWidget(
            self, simple=self._scan_progress_bar_simple, full=self._scan_progress_bar_full
        )

        # Bundle the progress bar with a separator
        separator = self._add_separator(separate_object=True)
        self._scan_progress_bar_with_separator = QWidget()
        self._scan_progress_bar_with_separator.layout = QHBoxLayout(
            self._scan_progress_bar_with_separator
        )
        self._scan_progress_bar_with_separator.layout.setContentsMargins(0, 0, 0, 0)
        self._scan_progress_bar_with_separator.layout.setSpacing(0)
        self._scan_progress_bar_with_separator.layout.addWidget(separator)
        self._scan_progress_bar_with_separator.layout.addWidget(self._scan_progress_hover)

        # Set Size
        self._scan_progress_bar_target_width = self.SCAN_PROGRESS_WIDTH
        self._scan_progress_bar_with_separator.setMaximumWidth(self._scan_progress_bar_target_width)

        self.status_bar.addWidget(self._scan_progress_bar_with_separator)

        # Visibility logic
        self._scan_progress_bar_with_separator.hide()
        self._scan_progress_bar_with_separator.setMaximumWidth(0)

        # Timer for hiding logic
        self._scan_progress_hide_timer = QTimer(self)
        self._scan_progress_hide_timer.setSingleShot(True)
        self._scan_progress_hide_timer.setInterval(self.STATUS_BAR_WIDGETS_EXPIRE_TIME)
        self._scan_progress_hide_timer.timeout.connect(self._animate_hide_scan_progress_bar)

        # Show / hide behaviour
        self._scan_progress_bar_simple.progress_started.connect(self._show_scan_progress_bar)
        self._scan_progress_bar_simple.progress_finished.connect(self._delay_hide_scan_progress_bar)

    def _show_scan_progress_bar(self):
        if self._scan_progress_hide_timer.isActive():
            self._scan_progress_hide_timer.stop()
        if self._scan_progress_bar_with_separator.isVisible():
            return

        # Make visible and reset width
        self._scan_progress_bar_with_separator.show()
        self._scan_progress_bar_with_separator.setMaximumWidth(0)

        self._show_container_anim = QPropertyAnimation(
            self._scan_progress_bar_with_separator, b"maximumWidth", self
        )
        self._show_container_anim.setDuration(300)
        self._show_container_anim.setStartValue(0)
        self._show_container_anim.setEndValue(self._scan_progress_bar_target_width)
        self._show_container_anim.setEasingCurve(QEasingCurve.OutCubic)
        self._show_container_anim.start()

    def _delay_hide_scan_progress_bar(self):
        """Start the countdown to hide the scan progress bar."""
        if hasattr(self, "_scan_progress_hide_timer"):
            self._scan_progress_hide_timer.start()

    def _animate_hide_scan_progress_bar(self):
        """Shrink container to the right, then hide."""
        self._hide_container_anim = QPropertyAnimation(
            self._scan_progress_bar_with_separator, b"maximumWidth", self
        )
        self._hide_container_anim.setDuration(300)
        self._hide_container_anim.setStartValue(self._scan_progress_bar_with_separator.width())
        self._hide_container_anim.setEndValue(0)
        self._hide_container_anim.setEasingCurve(QEasingCurve.InCubic)
        self._hide_container_anim.finished.connect(self._scan_progress_bar_with_separator.hide)
        self._hide_container_anim.start()

    def _add_separator(self, separate_object: bool = False) -> QWidget | None:
        """
        Add a vertically centred separator to the status bar or just return it as a separate object.
        """
        status_bar = self.statusBar()

        # The actual line
        line = QFrame()
        line.setFrameShape(QFrame.VLine)
        line.setFrameShadow(QFrame.Sunken)
        line.setFixedHeight(status_bar.sizeHint().height() - 2)

        # Wrapper to center the line vertically -> work around for QFrame not being able to center itself
        wrapper = QWidget()
        vbox = QVBoxLayout(wrapper)
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.addStretch()
        vbox.addWidget(line, alignment=Qt.AlignHCenter)
        vbox.addStretch()
        wrapper.setFixedWidth(line.sizeHint().width())

        if separate_object:
            return wrapper
        status_bar.addWidget(wrapper)

    def _init_bec_icon(self):
        icon = self.app.windowIcon()
        if icon.isNull():
            icon = QIcon()
            icon.addFile(
                os.path.join(MODULE_PATH, "assets", "app_icons", "bec_widgets_icon.png"),
                size=QSize(48, 48),
            )
            self.app.setWindowIcon(icon)

    def load_ui(self, ui_file):
        loader = UILoader(self)
        self.ui = loader.loader(ui_file)
        self.setCentralWidget(self.ui)

    def _fetch_theme(self) -> str:
        return self.app.theme.theme

    def _get_launcher_from_qapp(self):
        """
        Get the launcher from the QApplication instance.
        """
        from bec_widgets.applications.launch_window import LaunchWindow

        qapp = QApplication.instance()
        widgets = qapp.topLevelWidgets()
        widgets = [w for w in widgets if isinstance(w, LaunchWindow)]
        if widgets:
            return widgets[0]
        return None

    def _show_launcher(self):
        """
        Show the launcher if it exists.
        """
        launcher = self._get_launcher_from_qapp()
        if launcher:
            launcher.show()
            launcher.activateWindow()
            launcher.raise_()

    def _setup_menu_bar(self):
        """
        Setup the menu bar for the main window.
        """
        menu_bar = self.menuBar()

        ##########################################
        # Launch menu
        launch_menu = menu_bar.addMenu("New")

        open_launcher_action = QAction("Open Launcher", self)
        launch_menu.addAction(open_launcher_action)
        open_launcher_action.triggered.connect(self._show_launcher)

        ########################################
        # Theme menu
        theme_menu = menu_bar.addMenu("Theme")

        theme_group = QActionGroup(self)
        light_theme_action = QAction("Light Theme", self, checkable=True)
        dark_theme_action = QAction("Dark Theme", self, checkable=True)
        theme_group.addAction(light_theme_action)
        theme_group.addAction(dark_theme_action)
        theme_group.setExclusive(True)

        theme_menu.addAction(light_theme_action)
        theme_menu.addAction(dark_theme_action)

        # Connect theme actions
        light_theme_action.triggered.connect(lambda: self.change_theme("light"))
        dark_theme_action.triggered.connect(lambda: self.change_theme("dark"))

        # Set the default theme
        theme = self.app.theme.theme
        if theme == "light":
            light_theme_action.setChecked(True)
        elif theme == "dark":
            dark_theme_action.setChecked(True)

        ########################################
        # Help menu
        help_menu = menu_bar.addMenu("Help")

        help_icon = QApplication.style().standardIcon(QStyle.SP_MessageBoxQuestion)
        bug_icon = QApplication.style().standardIcon(QStyle.SP_MessageBoxInformation)

        bec_docs = QAction("BEC Docs", self)
        bec_docs.setIcon(help_icon)
        widgets_docs = QAction("BEC Widgets Docs", self)
        widgets_docs.setIcon(help_icon)
        bug_report = QAction("Bug Report", self)
        bug_report.setIcon(bug_icon)

        bec_docs.triggered.connect(BECWebLinksMixin.open_bec_docs)
        widgets_docs.triggered.connect(BECWebLinksMixin.open_bec_widgets_docs)
        bug_report.triggered.connect(BECWebLinksMixin.open_bec_bug_report)

        help_menu.addAction(bec_docs)
        help_menu.addAction(widgets_docs)
        help_menu.addAction(bug_report)

    ################################################################################
    # Status Bar Addons
    ################################################################################
    def display_app_id(self):
        """
        Display the app ID in the status bar.
        """
        if self.bec_dispatcher.cli_server is None:
            status_message = "Not connected"
        else:
            # Get the server ID from the dispatcher
            server_id = self.bec_dispatcher.cli_server.gui_id
            status_message = f"App ID: {server_id}"
        self._app_id_label.setText(status_message)

    @SafeSlot(dict, dict)
    def display_client_message(self, msg: dict, meta: dict):
        """
        Display a client message in the status bar.

        Args:
            msg(dict): The message to display, should contain:
            meta(dict): Metadata about the message, usually empty.
        """
        message = msg.get("message", "")
        expiration = msg.get("expire", 0)  # 0 → never expire
        self._client_info_label.setText(message)
        self._client_info_label_full.setText(message)

        # Restart the expiration timer if necessary
        if hasattr(self, "_client_info_expire_timer") and self._client_info_expire_timer.isActive():
            self._client_info_expire_timer.stop()
        if expiration and expiration > 0:
            self._client_info_expire_timer.start(int(expiration * 1000))

    ################################################################################
    # General and Cleanup Methods
    ################################################################################
    @SafeSlot(str)
    def change_theme(self, theme: str):
        """
        Change the theme of the application.

        Args:
            theme(str): The theme to apply, either "light" or "dark".
        """
        apply_theme(theme)

    def event(self, event):
        if event.type() == QEvent.Type.StatusTip:
            return True
        return super().event(event)

    def cleanup(self):
        central_widget = self.centralWidget()
        if central_widget is not None:
            central_widget.close()
            central_widget.deleteLater()
        if not isinstance(central_widget, BECWidget):
            # if the central widget is not a BECWidget, we need to call the cleanup method
            # of all widgets whose parent is the current BECMainWindow
            children = self.findChildren(BECWidget)
            for child in children:
                ancestor = WidgetHierarchy._get_becwidget_ancestor(child)
                if ancestor is self:
                    child.cleanup()
                    child.close()
                    child.deleteLater()

        # Timer cleanup
        if hasattr(self, "_client_info_expire_timer") and self._client_info_expire_timer.isActive():
            self._client_info_expire_timer.stop()
        if hasattr(self, "_scan_progress_hide_timer") and self._scan_progress_hide_timer.isActive():
            self._scan_progress_hide_timer.stop()

        ########################################
        # Status bar widgets cleanup

        # Client info label cleanup
        self._client_info_label.cleanup()
        self._client_info_hover.close()
        self._client_info_hover.deleteLater()
        # Scan progress bar cleanup
        self._scan_progress_bar_simple.close()
        self._scan_progress_bar_simple.deleteLater()
        self._scan_progress_bar_full.close()
        self._scan_progress_bar_full.deleteLater()
        self._scan_progress_hover.close()
        self._scan_progress_hover.deleteLater()
        super().cleanup()


class UILaunchWindow(BECMainWindow):
    RPC = True


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    main_window = UILaunchWindow()
    main_window.show()
    main_window.resize(800, 600)
    sys.exit(app.exec())



import sys

from qtpy.QtCore import QPoint, Qt
from qtpy.QtWidgets import QApplication, QHBoxLayout, QLabel, QProgressBar, QVBoxLayout, QWidget


class WidgetTooltip(QWidget):
    """Frameless, always-on-top window that behaves like a tooltip."""

    def __init__(self, content: QWidget) -> None:
        super().__init__(None, Qt.ToolTip | Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)
        self.setAttribute(Qt.WA_ShowWithoutActivating)
        self.setMouseTracking(True)
        self.content = content

        layout = QVBoxLayout(self)
        layout.setContentsMargins(6, 6, 6, 6)
        layout.addWidget(self.content)
        self.adjustSize()

    def leaveEvent(self, _event) -> None:
        self.hide()

    def show_above(self, global_pos: QPoint, offset: int = 8) -> None:
        self.adjustSize()
        screen = QApplication.screenAt(global_pos) or QApplication.primaryScreen()
        screen_geo = screen.availableGeometry()
        geom = self.geometry()

        x = global_pos.x() - geom.width() // 2
        y = global_pos.y() - geom.height() - offset

        x = max(screen_geo.left(), min(x, screen_geo.right() - geom.width()))
        y = max(screen_geo.top(), min(y, screen_geo.bottom() - geom.height()))

        self.move(x, y)
        self.show()


class HoverWidget(QWidget):

    def __init__(self, parent: QWidget | None = None, *, simple: QWidget, full: QWidget):
        super().__init__(parent)
        self._simple = simple
        self._full = full
        self._full.setVisible(False)
        self._tooltip = None

        lay = QVBoxLayout(self)
        lay.setContentsMargins(0, 0, 0, 0)
        lay.addWidget(simple)

    def enterEvent(self, event):
        # suppress empty-label tooltips for labels
        if isinstance(self._full, QLabel) and not self._full.text():
            return

        if self._tooltip is None:  # first time only
            self._tooltip = WidgetTooltip(self._full)
            self._full.setVisible(True)

        centre = self.mapToGlobal(self.rect().center())
        self._tooltip.show_above(centre)
        super().enterEvent(event)

    def leaveEvent(self, event):
        if self._tooltip and self._tooltip.isVisible():
            self._tooltip.hide()
        super().leaveEvent(event)

    def close(self):
        if self._tooltip:
            self._tooltip.close()
            self._tooltip.deleteLater()
            self._tooltip = None
        super().close()


################################################################################
# Demo
# Just a simple example to show how the HoverWidget can be used to display
# a tooltip with a full widget inside (two different widgets are used
# for the simple and full versions).
################################################################################


class DemoSimpleWidget(QLabel):  # pragma: no cover
    """A simple widget to be used as a trigger for the tooltip."""

    def __init__(self) -> None:
        super().__init__()
        self.setText("Hover me for a preview!")


class DemoFullWidget(QProgressBar):  # pragma: no cover
    """A full widget to be shown in the tooltip."""

    def __init__(self) -> None:
        super().__init__()
        self.setRange(0, 100)
        self.setValue(75)
        self.setFixedWidth(320)
        self.setFixedHeight(30)


if __name__ == "__main__":  # pragma: no cover
    app = QApplication(sys.argv)

    window = QWidget()
    window.layout = QHBoxLayout(window)
    hover_widget = HoverWidget(simple=DemoSimpleWidget(), full=DemoFullWidget())
    window.layout.addWidget(hover_widget)
    window.show()

    sys.exit(app.exec_())



from __future__ import annotations

from collections import defaultdict
from typing import Literal

import numpy as np
from bec_lib import bec_logger
from bec_lib.endpoints import MessageEndpoints
from pydantic import BaseModel, Field, field_validator
from qtpy.QtCore import Qt, QTimer
from qtpy.QtWidgets import QComboBox, QStyledItemDelegate, QWidget

from bec_widgets.utils import ConnectionConfig
from bec_widgets.utils.colors import Colors
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot
from bec_widgets.utils.toolbars.actions import NoCheckDelegate, WidgetAction
from bec_widgets.utils.toolbars.bundles import ToolbarBundle
from bec_widgets.widgets.control.device_input.base_classes.device_input_base import (
    BECDeviceFilter,
    ReadoutPriority,
)
from bec_widgets.widgets.control.device_input.device_combobox.device_combobox import DeviceComboBox
from bec_widgets.widgets.plots.image.image_base import ImageBase
from bec_widgets.widgets.plots.image.image_item import ImageItem

logger = bec_logger.logger


# noinspection PyDataclass
class ImageConfig(ConnectionConfig):
    color_map: str = Field(
        "plasma", description="The colormap  of the figure widget.", validate_default=True
    )
    color_bar: Literal["full", "simple"] | None = Field(
        None, description="The type of the color bar."
    )
    lock_aspect_ratio: bool = Field(
        False, description="Whether to lock the aspect ratio of the image."
    )

    model_config: dict = {"validate_assignment": True}
    _validate_color_map = field_validator("color_map")(Colors.validate_color_map)


class ImageLayerConfig(BaseModel):
    monitor: str | tuple | None = Field(None, description="The name of the monitor.")
    monitor_type: Literal["1d", "2d", "auto"] = Field("auto", description="The type of monitor.")
    source: Literal["device_monitor_1d", "device_monitor_2d", "auto"] = Field(
        "auto", description="The source of the image data."
    )


class Image(ImageBase):
    """
    Image widget for displaying 2D data.
    """

    PLUGIN = True
    RPC = True
    ICON_NAME = "image"
    USER_ACCESS = [
        # General PlotBase Settings
        "enable_toolbar",
        "enable_toolbar.setter",
        "enable_side_panel",
        "enable_side_panel.setter",
        "enable_fps_monitor",
        "enable_fps_monitor.setter",
        "set",
        "title",
        "title.setter",
        "x_label",
        "x_label.setter",
        "y_label",
        "y_label.setter",
        "x_limits",
        "x_limits.setter",
        "y_limits",
        "y_limits.setter",
        "x_grid",
        "x_grid.setter",
        "y_grid",
        "y_grid.setter",
        "inner_axes",
        "inner_axes.setter",
        "outer_axes",
        "outer_axes.setter",
        "auto_range_x",
        "auto_range_x.setter",
        "auto_range_y",
        "auto_range_y.setter",
        "minimal_crosshair_precision",
        "minimal_crosshair_precision.setter",
        # ImageView Specific Settings
        "color_map",
        "color_map.setter",
        "v_range",
        "v_range.setter",
        "v_min",
        "v_min.setter",
        "v_max",
        "v_max.setter",
        "lock_aspect_ratio",
        "lock_aspect_ratio.setter",
        "autorange",
        "autorange.setter",
        "autorange_mode",
        "autorange_mode.setter",
        "monitor",
        "monitor.setter",
        "enable_colorbar",
        "enable_simple_colorbar",
        "enable_simple_colorbar.setter",
        "enable_full_colorbar",
        "enable_full_colorbar.setter",
        "fft",
        "fft.setter",
        "log",
        "log.setter",
        "num_rotation_90",
        "num_rotation_90.setter",
        "transpose",
        "transpose.setter",
        "image",
        "main_image",
        "add_roi",
        "remove_roi",
        "rois",
    ]

    def __init__(
        self,
        parent: QWidget | None = None,
        config: ImageConfig | None = None,
        client=None,
        gui_id: str | None = None,
        popups: bool = True,
        **kwargs,
    ):
        if config is None:
            config = ImageConfig(widget_class=self.__class__.__name__)
        self.gui_id = config.gui_id
        self.subscriptions: defaultdict[str, ImageLayerConfig] = defaultdict(
            lambda: ImageLayerConfig(monitor=None, monitor_type="auto", source="auto")
        )
        super().__init__(
            parent=parent, config=config, client=client, gui_id=gui_id, popups=popups, **kwargs
        )
        self._init_toolbar_image()
        self.layer_removed.connect(self._on_layer_removed)
        self.scan_id = None

    ##################################
    ### Toolbar Initialization
    ##################################

    def _init_toolbar_image(self):
        """
        Initializes the toolbar for the image widget.
        """
        self.device_combo_box = DeviceComboBox(
            parent=self,
            device_filter=BECDeviceFilter.DEVICE,
            readout_priority_filter=[ReadoutPriority.ASYNC],
        )
        self.device_combo_box.addItem("", None)
        self.device_combo_box.setCurrentText("")
        self.device_combo_box.setToolTip("Select Device")
        self.device_combo_box.setFixedWidth(150)
        self.device_combo_box.setItemDelegate(NoCheckDelegate(self.device_combo_box))

        self.dim_combo_box = QComboBox(parent=self)
        self.dim_combo_box.addItems(["auto", "1d", "2d"])
        self.dim_combo_box.setCurrentText("auto")
        self.dim_combo_box.setToolTip("Monitor Dimension")
        self.dim_combo_box.setFixedWidth(100)
        self.dim_combo_box.setItemDelegate(NoCheckDelegate(self.dim_combo_box))

        self.toolbar.components.add_safe(
            "image_device_combo", WidgetAction(widget=self.device_combo_box, adjust_size=False)
        )
        self.toolbar.components.add_safe(
            "image_dim_combo", WidgetAction(widget=self.dim_combo_box, adjust_size=False)
        )

        bundle = ToolbarBundle("monitor_selection", self.toolbar.components)
        bundle.add_action("image_device_combo")
        bundle.add_action("image_dim_combo")

        self.toolbar.add_bundle(bundle)
        self.device_combo_box.currentTextChanged.connect(self.connect_monitor)
        self.dim_combo_box.currentTextChanged.connect(self.connect_monitor)

        crosshair_bundle = self.toolbar.get_bundle("image_crosshair")
        crosshair_bundle.add_action("image_autorange")
        crosshair_bundle.add_action("image_colorbar_switch")

        self.toolbar.show_bundles(
            [
                "monitor_selection",
                "plot_export",
                "mouse_interaction",
                "image_crosshair",
                "image_processing",
                "axis_popup",
            ]
        )

        QTimer.singleShot(0, self._adjust_and_connect)

    def _adjust_and_connect(self):
        """
        Adjust the size of the device combo box and populate it with preview signals.
        Has to be done with QTimer.singleShot to ensure the UI is fully initialized, needed for testing.
        """
        self._populate_preview_signals()
        self._reverse_device_items()
        self.device_combo_box.setCurrentText("")  # set again default to empty string

    def _populate_preview_signals(self) -> None:
        """
        Populate the device combo box with preview-signal devices in the
        format '<device>_<signal>' and store the tuple(device, signal) in
        the item's userData for later use.
        """
        preview_signals = self.client.device_manager.get_bec_signals("PreviewSignal")
        for device, signal, signal_config in preview_signals:
            label = signal_config.get("obj_name", f"{device}_{signal}")
            self.device_combo_box.addItem(label, (device, signal, signal_config))

    def _reverse_device_items(self) -> None:
        """
        Reverse the current order of items in the device combo box while
        keeping their userData and restoring the previous selection.
        """
        current_text = self.device_combo_box.currentText()
        items = [
            (self.device_combo_box.itemText(i), self.device_combo_box.itemData(i))
            for i in range(self.device_combo_box.count())
        ]
        self.device_combo_box.clear()
        for text, data in reversed(items):
            self.device_combo_box.addItem(text, data)
        if current_text:
            self.device_combo_box.setCurrentText(current_text)

    @SafeSlot()
    def connect_monitor(self, *args, **kwargs):
        """
        Connect the target widget to the selected monitor based on the current device and dimension.

        If the selected device is a preview-signal device, it will use the tuple (device, signal) as the monitor.
        """
        dim = self.dim_combo_box.currentText()
        data = self.device_combo_box.currentData()

        if isinstance(data, tuple):
            self.image(monitor=data, monitor_type="auto")
        else:
            self.image(monitor=self.device_combo_box.currentText(), monitor_type=dim)

    ################################################################################
    # Data Acquisition

    @SafeProperty(str)
    def monitor(self) -> str:
        """
        The name of the monitor to use for the image.
        """
        return self.subscriptions["main"].monitor or ""

    @monitor.setter
    def monitor(self, value: str):
        """
        Set the monitor for the image.

        Args:
            value(str): The name of the monitor to set.
        """
        if self.subscriptions["main"].monitor == value:
            return
        try:
            self.entry_validator.validate_monitor(value)
        except ValueError:
            return
        self.image(monitor=value)

    @property
    def main_image(self) -> ImageItem:
        """Access the main image item."""
        return self.layer_manager["main"].image

    ################################################################################
    # High Level methods for API
    ################################################################################
    @SafeSlot(popup_error=True)
    def image(
        self,
        monitor: str | tuple | None = None,
        monitor_type: Literal["auto", "1d", "2d"] = "auto",
        color_map: str | None = None,
        color_bar: Literal["simple", "full"] | None = None,
        vrange: tuple[int, int] | None = None,
    ) -> ImageItem | None:
        """
        Set the image source and update the image.

        Args:
            monitor(str): The name of the monitor to use for the image.
            monitor_type(str): The type of monitor to use. Options are "1d", "2d", or "auto".
            color_map(str): The color map to use for the image.
            color_bar(str): The type of color bar to use. Options are "simple" or "full".
            vrange(tuple): The range of values to use for the color map.

        Returns:
            ImageItem: The image object.
        """

        if self.subscriptions["main"].monitor:
            self.disconnect_monitor(self.subscriptions["main"].monitor)
        if monitor is None or monitor == "":
            logger.warning(f"No monitor specified, cannot set image, old monitor is unsubscribed")
            return None
        if isinstance(monitor, tuple):
            self.entry_validator.validate_monitor(monitor[0])
        else:
            self.entry_validator.validate_monitor(monitor)

        self.set_image_update(monitor=monitor, type=monitor_type)
        if color_map is not None:
            self.main_image.color_map = color_map
        if color_bar is not None:
            self.enable_colorbar(True, color_bar)
        if vrange is not None:
            self.vrange = vrange

        self._sync_device_selection()

        return self.main_image

    def _sync_device_selection(self):
        """
        Synchronize the device selection with the current monitor.
        """
        config = self.subscriptions["main"]
        if config.monitor is not None:
            for combo in (self.device_combo_box, self.dim_combo_box):
                combo.blockSignals(True)
            if isinstance(config.monitor, tuple):
                self.device_combo_box.setCurrentText(f"{config.monitor[0]}_{config.monitor[1]}")
            else:
                self.device_combo_box.setCurrentText(config.monitor)
            self.dim_combo_box.setCurrentText(config.monitor_type)
            for combo in (self.device_combo_box, self.dim_combo_box):
                combo.blockSignals(False)
        else:
            for combo in (self.device_combo_box, self.dim_combo_box):
                combo.blockSignals(True)
            self.device_combo_box.setCurrentText("")
            self.dim_combo_box.setCurrentText("auto")
            for combo in (self.device_combo_box, self.dim_combo_box):
                combo.blockSignals(False)

    ################################################################################
    # Post Processing
    ################################################################################

    @SafeProperty(bool)
    def fft(self) -> bool:
        """
        Whether FFT postprocessing is enabled.
        """
        return self.main_image.fft

    @fft.setter
    def fft(self, enable: bool):
        """
        Set FFT postprocessing.

        Args:
            enable(bool): Whether to enable FFT postprocessing.
        """
        self.main_image.fft = enable

    @SafeProperty(bool)
    def log(self) -> bool:
        """
        Whether logarithmic scaling is applied.
        """
        return self.main_image.log

    @log.setter
    def log(self, enable: bool):
        """
        Set logarithmic scaling.

        Args:
            enable(bool): Whether to enable logarithmic scaling.
        """
        self.main_image.log = enable

    @SafeProperty(int)
    def num_rotation_90(self) -> int:
        """
        The number of 90° rotations to apply counterclockwise.
        """
        return self.main_image.num_rotation_90

    @num_rotation_90.setter
    def num_rotation_90(self, value: int):
        """
        Set the number of 90° rotations to apply counterclockwise.

        Args:
            value(int): The number of 90° rotations to apply.
        """
        self.main_image.num_rotation_90 = value

    @SafeProperty(bool)
    def transpose(self) -> bool:
        """
        Whether the image is transposed.
        """
        return self.main_image.transpose

    @transpose.setter
    def transpose(self, enable: bool):
        """
        Set the image to be transposed.

        Args:
            enable(bool): Whether to enable transposing the image.
        """
        self.main_image.transpose = enable

    ################################################################################
    # Image Update Methods
    ################################################################################

    ########################################
    # Connections

    @SafeSlot()
    def set_image_update(self, monitor: str | tuple, type: Literal["1d", "2d", "auto"]):
        """
        Set the image update method for the given monitor.

        Args:
            monitor(str): The name of the monitor to use for the image.
            type(str): The type of monitor to use. Options are "1d", "2d", or "auto".
        """

        # TODO consider moving connecting and disconnecting logic to Image itself if multiple images
        if isinstance(monitor, tuple):
            device = self.dev[monitor[0]]
            signal = monitor[1]
            if len(monitor) == 3:
                signal_config = monitor[2]
            else:
                signal_config = device._info["signals"][signal]
            signal_class = signal_config.get("signal_class", None)
            if signal_class != "PreviewSignal":
                logger.warning(f"Signal '{monitor}' is not a PreviewSignal.")
                return

            ndim = signal_config.get("describe", None).get("signal_info", None).get("ndim", None)
            if ndim is None:
                logger.warning(
                    f"Signal '{monitor}' does not have a valid 'ndim' in its signal_info."
                )
                return

            if ndim == 1:
                self.bec_dispatcher.connect_slot(
                    self.on_image_update_1d, MessageEndpoints.device_preview(device.name, signal)
                )
                self.subscriptions["main"].source = "device_monitor_1d"
                self.subscriptions["main"].monitor_type = "1d"
            elif ndim == 2:
                self.bec_dispatcher.connect_slot(
                    self.on_image_update_2d, MessageEndpoints.device_preview(device.name, signal)
                )
                self.subscriptions["main"].source = "device_monitor_2d"
                self.subscriptions["main"].monitor_type = "2d"

        else:  # FIXME old monitor 1d/2d endpoint handling, present for backwards compatibility, will be removed in future versions
            if type == "1d":
                self.bec_dispatcher.connect_slot(
                    self.on_image_update_1d, MessageEndpoints.device_monitor_1d(monitor)
                )
                self.subscriptions["main"].source = "device_monitor_1d"
                self.subscriptions["main"].monitor_type = "1d"
            elif type == "2d":
                self.bec_dispatcher.connect_slot(
                    self.on_image_update_2d, MessageEndpoints.device_monitor_2d(monitor)
                )
                self.subscriptions["main"].source = "device_monitor_2d"
                self.subscriptions["main"].monitor_type = "2d"
            elif type == "auto":
                self.bec_dispatcher.connect_slot(
                    self.on_image_update_1d, MessageEndpoints.device_monitor_1d(monitor)
                )
                self.bec_dispatcher.connect_slot(
                    self.on_image_update_2d, MessageEndpoints.device_monitor_2d(monitor)
                )
                self.subscriptions["main"].source = "auto"
                logger.warning(
                    f"Updates for '{monitor}' will be fetch from both 1D and 2D monitor endpoints."
                )
                self.subscriptions["main"].monitor_type = "auto"

        logger.info(f"Connected to {monitor} with type {type}")
        self.subscriptions["main"].monitor = monitor

    def disconnect_monitor(self, monitor: str | tuple):
        """
        Disconnect the monitor from the image update signals, both 1D and 2D.

        Args:
            monitor(str|tuple): The name of the monitor to disconnect, or a tuple of (device, signal) for preview signals.
        """
        if isinstance(monitor, tuple):
            if self.subscriptions["main"].source == "device_monitor_1d":
                self.bec_dispatcher.disconnect_slot(
                    self.on_image_update_1d, MessageEndpoints.device_preview(monitor[0], monitor[1])
                )
            elif self.subscriptions["main"].source == "device_monitor_2d":
                self.bec_dispatcher.disconnect_slot(
                    self.on_image_update_2d, MessageEndpoints.device_preview(monitor[0], monitor[1])
                )
            else:
                logger.warning(
                    f"Cannot disconnect monitor {monitor} with source {self.subscriptions['main'].source}"
                )
                return
        else:  # FIXME old monitor 1d/2d endpoint handling, present for backwards compatibility, will be removed in future versions
            self.bec_dispatcher.disconnect_slot(
                self.on_image_update_1d, MessageEndpoints.device_monitor_1d(monitor)
            )
            self.bec_dispatcher.disconnect_slot(
                self.on_image_update_2d, MessageEndpoints.device_monitor_2d(monitor)
            )
        self.subscriptions["main"].monitor = None
        self._sync_device_selection()

    ########################################
    # 1D updates

    @SafeSlot(dict, dict)
    def on_image_update_1d(self, msg: dict, metadata: dict):
        """
        Update the image with 1D data.

        Args:
            msg(dict): The message containing the data.
            metadata(dict): The metadata associated with the message.
        """
        data = msg["data"]
        current_scan_id = metadata.get("scan_id", None)

        if current_scan_id is None:
            return
        if current_scan_id != self.scan_id:
            self.scan_id = current_scan_id
            self.main_image.clear()
            self.main_image.buffer = []
            self.main_image.max_len = 0
            if self.crosshair is not None:
                self.crosshair.reset()
        image_buffer = self.adjust_image_buffer(self.main_image, data)
        if self._color_bar is not None:
            self._color_bar.blockSignals(True)
        self.main_image.set_data(image_buffer)
        if self._color_bar is not None:
            self._color_bar.blockSignals(False)
        self.image_updated.emit()

    def adjust_image_buffer(self, image: ImageItem, new_data: np.ndarray) -> np.ndarray:
        """
        Adjusts the image buffer to accommodate the new data, ensuring that all rows have the same length.

        Args:
            image: The image object (used to store a buffer list and max_len).
            new_data (np.ndarray): The new incoming 1D waveform data.

        Returns:
            np.ndarray: The updated image buffer with adjusted shapes.
        """
        new_len = new_data.shape[0]
        if not hasattr(image, "buffer"):
            image.buffer = []
            image.max_len = 0

        if new_len > image.max_len:
            image.max_len = new_len
            for i in range(len(image.buffer)):
                wf = image.buffer[i]
                pad_width = image.max_len - wf.shape[0]
                if pad_width > 0:
                    image.buffer[i] = np.pad(wf, (0, pad_width), mode="constant", constant_values=0)
            image.buffer.append(new_data)
        else:
            pad_width = image.max_len - new_len
            if pad_width > 0:
                new_data = np.pad(new_data, (0, pad_width), mode="constant", constant_values=0)
            image.buffer.append(new_data)

        image_buffer = np.array(image.buffer)
        return image_buffer

    ########################################
    # 2D updates

    def on_image_update_2d(self, msg: dict, metadata: dict):
        """
        Update the image with 2D data.

        Args:
            msg(dict): The message containing the data.
            metadata(dict): The metadata associated with the message.
        """
        data = msg["data"]
        if self._color_bar is not None:
            self._color_bar.blockSignals(True)
        self.main_image.set_data(data)
        if self._color_bar is not None:
            self._color_bar.blockSignals(False)
        self.image_updated.emit()

    ################################################################################
    # Clean up
    ################################################################################

    @SafeSlot(str)
    def _on_layer_removed(self, layer_name: str):
        """
        Handle the removal of a layer by disconnecting the monitor.

        Args:
            layer_name(str): The name of the layer that was removed.
        """
        if layer_name not in self.subscriptions:
            return
        config = self.subscriptions[layer_name]
        if config.monitor is not None:
            self.disconnect_monitor(config.monitor)
            config.monitor = None

    def cleanup(self):
        """
        Disconnect the image update signals and clean up the image.
        """
        self.layer_removed.disconnect(self._on_layer_removed)
        for layer_name in list(self.subscriptions.keys()):
            config = self.subscriptions[layer_name]
            if config.monitor is not None:
                self.disconnect_monitor(config.monitor)
            del self.subscriptions[layer_name]
        self.subscriptions.clear()

        # Toolbar cleanup
        self.device_combo_box.close()
        self.device_combo_box.deleteLater()
        self.dim_combo_box.close()
        self.dim_combo_box.deleteLater()
        super().cleanup()


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication, QHBoxLayout

    app = QApplication(sys.argv)
    win = QWidget()
    win.setWindowTitle("Image Demo")
    ml = QHBoxLayout(win)

    image_popup = Image(popups=True)
    # image_side_panel = Image(popups=False)

    ml.addWidget(image_popup)
    # ml.addWidget(image_side_panel)

    win.resize(1500, 800)
    win.show()
    sys.exit(app.exec_())
